namespace Sink.Commands;

table ModifyEntity {
    revision: ulong;
    entityId: string;
    deletions: [string]; //A list of deleted properties
    domainType: string;
    delta: [ubyte]; //Contains an entity buffer with all changed properties set
    replayToSource: bool = true;
    modifiedProperties: [string];
    targetResource: string; //Contains the target resource for a move/copy operation
    removeEntity: bool = false; //This modification removes the entity
}

root_type ModifyEntity;
