namespace Sink.ApplicationDomain.Buffer;

table MailContact {
    name: string;
    email: string;
}

table Mail {
  uid:string;
  folder:string;
  sender:MailContact;
  to:[MailContact];
  cc:[MailContact];
  bcc:[MailContact];
  subject:string;
  date:string;
  unread:bool = false;
  important:bool = false;
  mimeMessage:string;
  draft:bool = false;
  trash:bool = false;
  sent:bool = false;
  messageId:string;
  parentMessageIds:[string];
  fullPayloadAvailable:bool = true;
}

root_type Mail;
file_identifier "AKFB";
