namespace Sink;

enum Operation : byte { Creation = 1, Modification, Removal }

table Metadata {
    revision: ulong;
    replayToSource: bool = true;
    operation: Operation = Modification;
    modifiedProperties: [string];
}

root_type Metadata;
