namespace Sink.ApplicationDomain.Buffer;

table ContactEmail {
    type: int;
    email: string;
}

table Contact {
  uid:string;
  fn:string;
  firstname:string;
  lastname:string;
  addressbook:string;
  emails: [ContactEmail];
  vcard: string;
  photo: string;
}

root_type Contact;
file_identifier "AKFB";
