namespace Sink.ApplicationDomain.Buffer;

table Event {
  uid:string;
  summary:string;
  description:string;
  startTime:string;
  endTime:string;
  allDay:bool = false;
  recurring:bool = false;
  ical:string;
  calendar:string;
}

root_type Event;
file_identifier "AKFB";
